--[[
    Aaron MacDonald
    August 31, 2010
]]

require('utility')
require('levels')

collision = {}

local function moveRect(radius, x, y, dx, dy)
    local width = (radius * 2) + math.abs(dx)
    local height = (radius * 2) + math.abs(dy)

    local rx = x - radius
    if dx < 0 then
        rx = rx + dx
    end

    local ry = y - radius
    if dy < 0 then
        ry = ry + dy
    end

    return utility.rectangle(rx, ry, width, height)
end

----------------------------------------------------------------------

-- Lines go counter-clockwise

local function tileInfo(tile)
    local px = levels.pixel(tile.x)
    local py = levels.pixel(tile.y)

    return px, py
end

local function northSide(tile)
    local px, py = tileInfo(tile)

    -- Runs east to west
    return { x = px + levels.TILE_SIZE,
             y = py,
             dx = -levels.TILE_SIZE,
             dy = 0
           }
end

local function southSide(tile)
    local px, py = tileInfo(tile)

    -- Runs west to east
    return { x = px,
             y = py + levels.TILE_SIZE,
             dx = levels.TILE_SIZE,
             dy = 0
           }
end

local function eastSide(tile)
    local px, py = tileInfo(tile)

    -- Runs south to north
    return { x = px + levels.TILE_SIZE,
             y = py + levels.TILE_SIZE,
             dx = 0,
             dy = -levels.TILE_SIZE
           }
end

local function westSide(tile)
    local px, py = tileInfo(tile)

    -- Runs north to south
    return { x = px,
             y = py,
             dx = 0,
             dy = levels.TILE_SIZE
           }
end

----------------------------------------------------------------------

local function dotProduct(ax, ay, bx, by)
    return (ax * bx) + (ay * by)
end

----------------------------------------------------------------------

local function movingToVerticalLine(act, lineX)
    local dx = lineX - act.x
    local dy = 0

    return dotProduct(act.dx, act.dy, dx, dy)
end

local function movingToHorizontalLine(act, lineY)
    local dx = 0
    local dy = lineY - act.y

    return dotProduct(act.dx, act.dy, dx, dy)
end

----------------------------------------------------------------------

local function testNorthSide(tile, act, testingCorners)
    local result = false

    local line = northSide(tile)
    local lineY = line.y

    if not testingCorners then
        lineY = lineY - act.type.radius
    end

    if act.y <= lineY then
        result = movingToHorizontalLine(act, line.y)
    end

    return result
end

local function testSouthSide(tile, act, testingCorners)
    local result = false

    local line = southSide(tile)
    local lineY = line.y

    if not testingCorners then
        lineY = lineY + act.type.radius
    end

    if act.y >= lineY then
        result = movingToHorizontalLine(act, line.y)
    end

    return result
end

local function testEastSide(tile, act, testingCorners)
    local result = false

    local line = eastSide(tile)
    local lineX = line.x

    if not testingCorners then
        lineX = lineX + act.type.radius
    end

    if act.x >= lineX then
        result = movingToVerticalLine(act, line.x)
    end

    return result
end

local function testWestSide(tile, act, testingCorners)
    local result = false

    local line = westSide(tile)
    local lineX = line.x

    if not testingCorners then
        lineX = lineX - act.type.radius
    end

    if act.x <= lineX then
        result = movingToVerticalLine(act, line.x)
    end

    return result
end

----------------------------------------------------------------------

-- shiftX and shiftY is direction to shift line to account for
-- actor's radius
local function checkWall(act, tile, sideCheck, sideLine,
  shiftX, shiftY, sidesToTest, cornersToTest)
    local line = sideLine(tile)

    if sideCheck(tile, act) then
        -- Sides
        table.insert(sidesToTest, {line = line, shiftX = shiftX,
          shiftY = shiftY})
    end

    if sideCheck(tile, act, true) then
        -- Corners

        local cornerStart = {x = line.x, y = line.y}
        local cornerEnd = {x = line.x + line.dx, y = line.y + line.dy}

        -- Check for duplicates
        for _, c in ipairs(cornersToTest) do
            if cornerStart and ((c.x == cornerStart.x)
              and (c.y == cornerStart.y)) then
                cornerStart = nil
            elseif cornerEnd
              and ((c.x == cornerEnd.x) and (c.y == cornerEnd.y)) then
                cornerEnd = nil
            end
        end

        if cornerStart then
            table.insert(cornersToTest, cornerStart)
        end
        if cornerEnd then
            table.insert(cornersToTest, cornerEnd)
        end
    end
end

----------------------------------------------------------------------

local function lineIntersectionTime(act, line)
    local result

    -- Vector from line start to actor
    local dx = act.x - line.x
    local dy = act.y - line.y

    local d = (line.dy * act.dx) - (line.dx * act.dy)
    local nC = (line.dx * dy) - (line.dy * dx)
    local nL = (act.dx * dy) - (act.dy * dx)

    local timeC = nC / d
    local timeL = nL / d

    if (timeL >= 0) and (timeL <= 1) and (timeC >= 0) then
        -- Actor's path crosses side sometime in the future
        result = timeC
    end

    return result
end

local function checkSideIntersection(act, wall, remainingTime)
    local resultTime = remainingTime
    local realDX, realDY
    local newDX, newDY

    local wallLine = wall.line
    wallLine.x = wallLine.x + (wall.shiftX * act.type.radius)
    wallLine.y = wallLine.y + (wall.shiftY * act.type.radius)

    local testTime = lineIntersectionTime(act, wallLine)

    if testTime and (testTime < remainingTime) then
        -- Actor's path intersects wall, and does so before end of
        -- maximum frame time
        resultTime = testTime

        realDX = utility.floor2(act.dx * resultTime)
        realDY = utility.floor2(act.dy * resultTime)

        -- Project actor's velocity vector onto wall vector to get its
        -- new velocity

        local wallLenSqr = levels.TILE_SIZE * levels.TILE_SIZE
        local dotprod = dotProduct(act.dx, act.dy,
          wallLine.dx, wallLine.dy)
        local m = dotprod / wallLenSqr

        newDX = utility.floor2(m * wallLine.dx)
        newDY = utility.floor2(m * wallLine.dy)
    end

    return resultTime, realDX, realDY, newDX, newDY
end

local function checkCorner(act, corner, remainingTime)
    local resultTime = remainingTime
    local realDX, realDY
    local newDX, newDY

    local radius = act.type.radius

    -- Vector from actor to corner
    local cornDX = corner.x - act.x
    local cornDY = corner.y - act.y

    local cornLenSqr = (cornDX * cornDX) + (cornDY * cornDY)
    local velocitySqr = (act.dx * act.dx) + (act.dy * act.dy)

    local lenSqr = math.sqrt(cornLenSqr) - radius -- Not squared yet
    lenSqr = lenSqr * lenSqr -- Here we go

    if lenSqr < velocitySqr then
        -- Project vector from actor to corner onto actor's velocity
        -- vector.  This will get the closest point actor will ever
        -- reach corner.

        -- Velocity unit vector
        local velocity = math.sqrt(velocitySqr)
        local vnx = act.dx / velocity
        local vny = act.dy / velocity

        -- Length of the projection vector
        local dotprod = dotProduct(vnx, vny, cornDX, cornDY)

        if dotprod > 0 then
            -- Actor-to-corner vector and projection vector form a
            -- right triangle (actor-to-corner vector is the
            -- hypotenuse)

            local minDistSqr = cornLenSqr - (dotprod * dotprod)

            local radiusSqr = radius * radius

            if minDistSqr < radiusSqr then
                -- The min distance forms a right triangle with a
                -- hypotenuse with length equal to the radius.
                -- Find the length of the other side.

                local travOffsetSqr = radiusSqr - minDistSqr

                -- Distance actor travels before hitting corner
                local realDist = dotprod - math.sqrt(travOffsetSqr)

                if realDist < velocity then
                    local testTime = realDist / velocity

                    if testTime < remainingTime then
                        resultTime = testTime

                        realDX = utility.floor2(act.dx * resultTime)
                        realDY = utility.floor2(act.dy * resultTime)

                        local fx = act.x + realDX
                        local fy = act.y + realDY

                        -- Vector from corner to actor's final
                        -- position
                        local fdx = fx - corner.x
                        local fdy = fy - corner.y

                        -- Right hand normal
                        local normalX = -fdy
                        local normalY = fdx

                        -- Project actor's velocity vector onto
                        -- tangent of corner

                        local dotprod = dotProduct(act.dx, act.dy,
                          normalX, normalY)
                        local m = dotprod / radiusSqr

                        newDX = utility.floor2(m * normalX)
                        newDY = utility.floor2(m * normalY)
                    end
                end
            end
        end
    end

    return resultTime, realDX, realDY, newDX, newDY
end

----------------------------------------------------------------------

local function actorVSTile(tile, act, remainingTime)
    local sidesToTest = {} -- Each element contains wall line and
                           -- vector to shift line by to move towards
                           -- circle when accounting for radius
    local cornersToTest = {}

    checkWall(act, tile, testNorthSide, northSide, 0, -1, sidesToTest,
      cornersToTest)
    checkWall(act, tile, testSouthSide, southSide, 0, 1, sidesToTest,
      cornersToTest)
    checkWall(act, tile, testEastSide, eastSide, 1, 0, sidesToTest,
      cornersToTest)
    checkWall(act, tile, testWestSide, westSide, -1, 0, sidesToTest,
      cornersToTest)

    -- Find first collision

    local resultTime = remainingTime
    local realDX, realDY
    local newDX, newDY

    local testTime, testRDX, testRDY, testNDX, testNDY

    -- Check sides
    for _, wall in ipairs(sidesToTest) do
        testTime, testRDX, testRDY, testNDX, testNDY
          = checkSideIntersection(act, wall, resultTime)
        if testTime < resultTime then
            resultTime = testTime
            realDX = testRDX
            realDY = testRDY
            newDX = testNDX
            newDY = testNDY
        end
    end

    -- Check corners
    for _, corner in ipairs(cornersToTest) do
        testTime, testRDX, testRDY, testNDX, testNDY = checkCorner(
          act, corner, resultTime)
        if testTime < resultTime then
            resultTime = testTime
            realDX = testRDX
            realDY = testRDY
            newDX = testNDX
            newDY = testNDY
        end
    end

    return resultTime, realDX, realDY, newDX, newDY
end

function collision.actorVSLevel(level, act, remainingTime)
    local radius = act.type.radius
    local moveRect = moveRect(radius, act.x, act.y, act.dx, act.dy)

    local tiles = levels.tilesInRect(level, moveRect,
      function (x, y) return levels.isWall(level, x, y) end)

    -- Find first collision

    local resultTime = remainingTime
    local realDX = act.dx
    local realDY = act.dy
    local newDX, newDY

    local testTime, testRDX, testRDY, testNDX, testNDY
    for _, t in ipairs(tiles) do
        testTime, testRDX, testRDY, testNDX, testNDY = actorVSTile(
          t, act, resultTime)

        if testTime < resultTime then
            resultTime = testTime
            realDX = testRDX
            realDY = testRDY
            newDX = testNDX
            newDY = testNDY
        end
    end

    return resultTime, realDX, realDY, newDX, newDY
end