--[[
    Aaron MacDonald
    August 28, 2010
]]

require('graphics')

local LEVEL_DIR = 'levels/'
local LEVEL_EXT = '.lua'

local LAYER_GROUND = 'Ground'
local LAYER_OVER = 'Overhang'
local LAYER_ACTORS = 'Actors'

local DATA_PLAYER = 'Player'

levels = {}
levels.TILE_SIZE = 64

local TILE_HALF_SIZE = math.floor(levels.TILE_SIZE / 2)

----------------------------------------------------------------------

local function loadLevelLayer(layer, width, index)
    for nx = 0, width - 1 do
        layer[nx] = {}
    end

    local maxTileID = 0

    local x
    local y = -1

    for i, tile in ipairs(map.layers[index].data) do
        if ((i - 1) % width) == 0 then
            x = 0
            y = y + 1
        end

        layer[x][y] = tile.gid

        x = x + 1
        
        if maxTileID < tile.gid then
            maxTileID = tile.gid
        end
    end

    return maxTileID
end

----------------------------------------------------------------------

function levels.indexToCoords(index, width)
    -- Index starts at 1, because Lua's array indexing is bull#$*%.
    local realI = index - 1
    local y = math.floor(realI / width)
    local x = realI - (y * width)

    return x, y
end

--[[
function levels.coordsToIndex(x, y, width)
    -- Index starts at one
    return (y * width) + x + 1
end

function levels.maxIndex(level)
    return level.width * level.height
end
]]

function levels.inBounds(level, x, y)
    return (x >= 0) and (y >= 0)
      and (x < level.width) and (y < level.height)
end

function levels.pixel(coord)
    return coord * levels.TILE_SIZE
end

function levels.tile(pixel)
    return math.floor(pixel / levels.TILE_SIZE)
end

function levels.pixelWidth(level)
    return level.width * levels.TILE_SIZE
end

function levels.pixelHeight(level)
    return level.height * levels.TILE_SIZE
end

function levels.offsetPixel(pixel)
    return pixel - (levels.tile(pixel) * levels.TILE_SIZE)
end

function levels.tilesInRect(level, rect, filter)
    local minX = levels.tile(rect.x)
    local minY = levels.tile(rect.y)
    local maxX = levels.tile(rect.x + rect.width - 1)
    local maxY = levels.tile(rect.y + rect.height - 1)

    if minX < 0 then
        minX = 0
    end

    if minY < 0 then
        minY = 0
    end

    if maxX >= level.width then
        maxX = level.width - 1
    end

    if maxY >= level.height then
        maxY = level.height - 1
    end

    local result = {}

    for x = minX, maxX do
        for y = minY, maxY do
            if (not filter) or filter(x, y) then
                table.insert(result, {x = x, y = y})
            end
        end
    end

    return result
end

function levels.isWall(level, x, y)
    return level.wallTiles[level.groundLayer[x][y]]
end

----------------------------------------------------------------------

function levels.newLevel(file, cameraX, cameraY)
    local level = {}

    -- Because Tiled's Lua exporter is stupid, I am not allowed to
    -- name something 'map' anywhere else in my code.

    local levelChunk = love.filesystem.load(LEVEL_DIR .. file
      .. LEVEL_EXT)
    levelChunk()

    -- Load properties
    level.minY = 0
    if map.properties then
        level.minY = map.properties.minY
    end

    -- Load terrain

    local layerIndex = 1

    level.width = map.width
    level.height = map.height

    assert(map.layers[layerIndex].name == LAYER_GROUND)

    level.groundLayer = {}
    local maxTileID = loadLevelLayer(level.groundLayer, level.width,
      layerIndex)

    layerIndex = layerIndex + 1
    if map.layers[layerIndex].name == LAYER_OVER then

        level.overLayer = {}
        local newMaxTileID = loadLevelLayer(level.overLayer,
          level.width, layerIndex)

        if newMaxTileID > maxTileID then
            maxTileID = newMaxTileID
        end

        layerIndex = layerIndex + 1
    end

    -- Load actor data

    if map.layers[layerIndex].name == LAYER_ACTORS then
        for _, data in ipairs(map.layers[layerIndex]) do
            if data.type == DATA_PLAYER then
                -- Place in center of tile
                local playerX = data.x + TILE_HALF_SIZE
                local playerY = data.y + TILE_HALF_SIZE

                level.playerSpawn = {x = playerX, y = playerY}
            end
        end
    end

    -- Load tileset

    assert(#map.tilesets == 1)
    assert(#map.tilesets[1] > 1)

    level.wallTiles = {}

    for i = 2, #map.tilesets[1] do
        local tileType = map.tilesets[1][i]
        assert(tileType.label == 'tile')

        if tileType.properties.wall then
            level.wallTiles[tileType.id + 1] = true
        end
    end

    local tilesetName = map.tilesets[1][1].source
    local tileset = graphics.loadImage(tilesetName, true)

    local tilesetWidth = tileset:getWidth()
    local tilesetHeight = tileset:getHeight()

    local tilesetTileWidth = math.floor(
      tilesetWidth / levels.TILE_SIZE)

    level.tilesetQuads = {}
    for i = 1, maxTileID do
        local x, y = levels.indexToCoords(i, tilesetTileWidth)
        x = x * levels.TILE_SIZE
        y = y * levels.TILE_SIZE

        local quad = love.graphics.newQuad(x, y,
          levels.TILE_SIZE, levels.TILE_SIZE,
          tilesetWidth, tilesetHeight)

        table.insert(level.tilesetQuads, quad)
    end

    level.tileset = tileset

    --[[level.groundBatch = love.graphics.newSpriteBatch(tileset,
      screenTotalTiles)
    level.overBatch = love.graphics.newSpriteBatch(tileset,
      screenTotalTiles)]]

    map = nil
    return level
end

----------------------------------------------------------------------

local function drawLayer(level, layer, camera)
    local camTileX = levels.tile(camera.x)
    local camTileY = levels.tile(camera.y)

    local camTileWidth = math.ceil(camera.width / levels.TILE_SIZE)

    if ((camera.width % levels.TILE_SIZE) > 0)
      or (levels.offsetPixel(camera.x) > 0) then
        camTileWidth = camTileWidth + 1
    end

    local camTileHeight = math.ceil((camera.height + level.minY)
      / levels.TILE_SIZE)

    if ((camera.height % levels.TILE_SIZE) > 0)
      or (levels.offsetPixel(camera.y) > 0) then
        camTileHeight = camTileHeight + 1
    end

    -- Stay within map

    if camTileX < 0 then
        camTileWidth = camTileWidth + camTileX
        camTileX = 0
    end

    if camTileY < 0 then
        camTileHeight = camTileHeight + camTileY
        camTileY = 0
    end

    if (camTileX + camTileWidth) >= level.width then
        camTileWidth = level.width - camTileX
    end

    if (camTileY + camTileHeight) >= level.height then
        camTileHeight = level.height - camTileY
    end

    -- Draw the map

    for x = camTileX, camTileX + camTileWidth - 1 do
        for y = camTileY, camTileY + camTileHeight - 1 do
            local tileID = layer[x][y]
            if tileID > 0 then
                local quad = level.tilesetQuads[tileID]
                love.graphics.drawq(level.tileset, quad,
                  x * levels.TILE_SIZE, y * levels.TILE_SIZE)
            end
        end
    end
end

function levels.drawGround(level, camera)
    drawLayer(level, level.groundLayer, camera)
end

function levels.drawOver(level, camera)
    if level.overLayer then
        drawLayer(level, level.overLayer, camera)
    end
end